/* to do:
 *    make beam current timing configurable
 *    make FPGA respond to sequential I/Q, not parallel
 *    implement "dump setup" option
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>  /* for stdrup() */
#include <math.h>    /* for sqrt() */
#include <unistd.h>  /* for getopt(), and getpid() used as random number seed */

/* Prefix used to avoid name space collisions with environment variables
 * possibly used by other programs.  You should be able to override this
 * during compilation with "-DPFX=FOO_" or the equivalent for your compiler.
 */
#ifndef PFX
#define PFX "OP_"
#endif

/* Maybe I should have used c++ with a complex class library.
 * Then again, maybe not. */
typedef struct {
	float re;
	float im;
} cpx;

/* ans must not overlap either x or y */
void c_mult(cpx *ans, const cpx *x, const cpx *y)
{
	ans->re = x->re*y->re - x->im*y->im;
	ans->im = x->im*y->re + x->re*y->im;
}

/* works even if ans and x overlap */
void c_inv(cpx *ans, const cpx *x)
{
	double r2 = x->re*x->re + x->im*x->im;
	ans->re =  x->re / r2;
	ans->im = -x->im / r2;
}

struct filter_element {
	cpx coeff;
	cpx a;
	cpx b;
	cpx state;
};

typedef struct {
	int order;
	double scale;
	double dt;
	cpx prev_in;
	struct filter_element s[1];  /* real size determined at run time */
} filter;

void step_filter(filter *f, cpx *input, cpx *output)
{
	int i;
	cpx prev;
	cpx avo, bvi;
	prev.re = 0.5*(input->re+f->prev_in.re);
	prev.im = 0.5*(input->im+f->prev_in.im);
	for (i=0; i<f->order; i++) {
		c_mult(&avo, &(f->s[i].a), &(f->s[i].state));
		c_mult(&bvi, &(f->s[i].b), &prev);
		prev.re = f->s[i].state.re;
		prev.im = f->s[i].state.im;
		f->s[i].state.re = avo.re + bvi.re;
		f->s[i].state.im = avo.im + bvi.im;
		prev.re = 0.5*(prev.re + f->s[i].state.re);
		prev.im = 0.5*(prev.im + f->s[i].state.im);
	}
	if (output) {
		*output = f->s[f->order-1].state;
		output->re *= f->scale;
		output->im *= f->scale;
	}
	f->prev_in = *input;
}

filter *new_filter(double dt, char *setup)
{
	char *p=setup, *newp;
	double scale=1.0;
	int i;
	cpx num, den;
	filter *f=NULL;
	/* fprintf(stderr,"new_filter()\n"); */
	for (i=0; ; i++) {
		double re, im;
		re = strtod(p, &newp);
		if (newp == p || *newp != ',') break;
		p = newp+1;
		im = strtod(p, &newp);
		if (newp == p) break;
		f=realloc(f,sizeof(filter)+
			(i)*sizeof(struct filter_element));
		if (f==NULL) {
			perror("realloc");
			return f;
		}
		/* fprintf(stderr,"   term %d: %f %f\n",i,re,im); */
		f->s[i].coeff.re=re;
		f->s[i].coeff.im=im;
		f->s[i].state.re=0.0;
		f->s[i].state.im=0.0;
		scale *= sqrt(re*re+im*im);

		/* a = (1+0.5*dt*p)/(1-0.5*dt*p); */
		den.re=1-0.5*dt*re;
		den.im= -0.5*dt*im;
		c_inv(&den,&den);
		num.re=1+0.5*dt*re;
		num.im= +0.5*dt*im;
		c_mult(&(f->s[i].a),&num,&den);

		/* b = dt/(1-0.5*dt*p); */
		num.re=dt;
		num.im=0.0;
		c_mult(&(f->s[i].b),&num,&den);
		p=newp;
	}
	f->order=i;
	f->scale=scale;
	f->prev_in.re=0;
	f->prev_in.im=0;
	f->dt = dt;
	return f;
}

typedef struct {
	int order;
	int now;
	cpx circle[1];   /* real size determined at run time */
} delay;

void step_delay(delay *d, cpx *input, cpx *output)
{
	int n = d->now;
	cpx result;
	if (d->order == 0) {
		result = *input;
	} else {
		result = d->circle[n];
		d->circle[n] = *input;
		++n;
		if (n >= d->order) n=0;
		d->now = n;
	}
	if (output) *output = result;
}

delay *new_delay(int order)
{
	delay *d=calloc(1, sizeof(delay)+(order-1)*sizeof(cpx));
	if (d) {
		d->order = order;
		d->now = 0;
	}
	return d;
}

void saturate(cpx *v, double sharpness)
{
	double mag = sqrt(v->re*v->re+v->im*v->im);
	double scale = (sharpness==0)?1.0:pow(1.0+pow(mag,sharpness),-1./sharpness);
	/* printf("scale=%f, outmag=%f\n", scale, scale*mag); */
	v->re *= scale;
	v->im *= scale;
}

/* output has mean of zero and standard deviation of 1.0
 * not really normal distribution, but close;  peak/rms is 3.0  */
double grand(void)
{
	double g=0.0;
	int i;
	for (i=0; i<3; i++) {
		/* each term of the sum is uniform over [-.5,5] */
		g += ((double) rand()/ (double) RAND_MAX - 0.5);
	}
	return g*2.0;
}

static struct {filter *trf1, *trf2, *cav, *rxf; delay *del1, *del2;
		int npt, t1, t2; double dt, igrl, clip, nois;
		cpx beam, drv1, drv2, drv3, gain, setp;}
	op;

void advance(int t, cpx *xmit, cpx *be, cpx *recv)
{
	cpx o1, o2, o3, o4, o5, o6;

	printf("%f %f %f ", t*op.dt, xmit->re, xmit->im);
	step_delay(op.del1, xmit, &o1);
	step_filter(op.trf1, &o1, &o2);
	saturate(&o2,op.clip);
	step_filter(op.trf2, &o2, &o3);
	printf("%f %f ", o3.re, o3.im);

	/* add beam loading */
	o3.re += be->re;
	o3.im += be->im;
	step_filter(op.cav, &o3, &o4);
	printf("%f %f ", o4.re, o4.im);

	step_delay(op.del2, &o4, &o5);
	step_filter(op.rxf, &o5, &o6);
	
	printf("%f %f\n", o6.re, o6.im);

	/* ADC and preamp noise */
	o6.re += op.nois * grand();
	o6.im += op.nois * grand();
	if (recv) *recv = o6;
}

char *lookup(char *name, char *dflt)
{
	char *s = getenv(name);
	if (!s) s=dflt;
	return s;
}

char *set_cpx(cpx *x, char *p)
{
	char *newp;
	if (x == NULL) return p;
	x->re = strtod(p, &newp);
	if (newp != p && *newp == ',') {
		x->im = strtod(newp+1, &newp);
	} else {
		x->im = 0.0;
	}
	return newp;
}

void initialize(void)
{
	double dt;
	op.dt   = dt =      atof(lookup(PFX "DT"  , "0.025"));              /* digital sample rate */
	op.npt  =           atoi(lookup(PFX "NPT" , "2900"));               /* number of sample points */
	op.del1 = new_delay(atoi(lookup(PFX "DEL1", "17")));                /* cables plus triode */
	op.trf1 = new_filter(dt, lookup(PFX "TRF1", "-7.5,2.7 -7.5,-2.7")); /* triode driver filter */
	op.trf2 = new_filter(dt, lookup(PFX "TRF2", "-9,-0.0"));            /* triode output filter */
	op.cav  = new_filter(dt, lookup(PFX "CAV" , "-.075,+.010"));        /* cavity filter */
	op.del2 = new_delay(atoi(lookup(PFX "DEL2", "12")));                /* cables plus ADC */
	op.rxf  = new_filter(dt, lookup(PFX "RXF" , "-18,0 -9.5,-15 -9.5,15")); /* receive 50 MHz filter */
	op.nois =           atof(lookup(PFX "NOIS", "0.003"));              /* rms ADC noise */
	set_cpx(&op.beam,        lookup(PFX "BEAM", "0,0.10"));             /* beam current vector */
	op.clip =           atof(lookup(PFX "CLIP", "5.0"));                /* triode clip harshness */
	op.t1   =           atoi(lookup(PFX "T1"  , "750"));                /* length of initial ramp */
	op.t2   =           atoi(lookup(PFX "T2"  , "1410"));               /* time of feedforward */
	set_cpx(&op.drv1,        lookup(PFX "DRV1", "1.50,0.00"));          /* initial drive vector */
	set_cpx(&op.drv2,        lookup(PFX "DRV2", "0.80,0.00"));          /* second drive vector */
	set_cpx(&op.drv3,        lookup(PFX "DRV3", "0.00,-0.05"));         /* feedforward vector */
	set_cpx(&op.setp,        lookup(PFX "SETP", "0.74,0.0"));           /* feedback setpoint */
	set_cpx(&op.gain,        lookup(PFX "GAIN", "-6.93,-1.0"));         /* feedback proportional gain */
	op.igrl =           atof(lookup(PFX "IGRL", "0.002"));              /* feedback integral term */
}

/* SNS beam timing model -- hard coded for 950 ns period and 25 ns time steps,
 * includes ramp from 12 to 24 ticks pulse width (31% to 63% duty factor) */
int beam_on(int t)
{
	int p, on, cycle, tmin, tmax, cstart=25;
	p=t%38;
	cycle=t/38;
	tmin=(38-cycle+cstart)/2;      if (tmin< 6) tmin= 6;
	tmax=(38+cycle-cstart)/2;      if (tmax>32) tmax=32;
	on = cycle>(cstart+12) && p>tmin && p<tmax;
	/* don't worry about the end of the beam pulse yet. */
	return on;
}

void fpga(int t, cpx *recv, cpx *xmit)
{
	cpx error, corr;
	static cpx state;

	if (  t==0   ) { state = op.drv2; }
	if ( t<op.t1 ) { xmit->re = op.drv1.re; xmit->im = op.drv1.im; return;}

	/* "feedforward" */
	if ( t==op.t2 ) { state.re += op.drv3.re;  state.im += op.drv3.im; }

	error.re = recv->re - op.setp.re;
	error.im = recv->im - op.setp.im;
	c_mult(&corr, &error, &op.gain);
	state.re += op.igrl * corr.re;
	state.im += op.igrl * corr.im;
	xmit->re = corr.re + state.re;
	xmit->im = corr.im + state.im;
}

/* Just so WinXX users don't have to fiddle with environment variables,
 * which I am told is a pain. */
void read_config_file(const char *name)
{
	char buff[100];
	FILE *f=fopen(name,"r");
	while (fgets(buff,sizeof(buff),f)) {
		if (buff[0] == '#') continue;
		if (putenv(strdup(buff))<0) {
			perror(buff);
			exit(1);
		};
	}
}

void usage(char *s)
{
	fprintf(stderr,"Usage: %s [-r rand_seed]\n"
	"\tfull configuration from environment variables\n",s);
	exit(1);
}

int main(int argc, char *argv[])
{
	cpx dac, beam, recv;
	int opt, i;
	int rand_seed=getpid();

	while ((opt=getopt(argc, argv, "r:f:")) != -1) {
		switch (opt) {
		case 'f':
			read_config_file(optarg);
			break;
		case 'r':
			rand_seed = atoi(optarg);
			break;
		default:
			usage(argv[0]);
		}
	}
	if (optind != argc) usage(argv[0]);  /* no valid non-option arguments */
	srand(rand_seed);
	initialize();
	recv.re=0; recv.im=0;  /* control loop has no input on first cycle */
	for (i=0; i<op.npt; i++) {
		/* control system drive */
		fpga(i, &recv, &dac);
		/* beam current, "normalized", see model.tex */
		if (beam_on(i)) { beam = op.beam; }
		           else { beam.re = 0.0; beam.im = 0.0; }
		advance(i, &dac, &beam, &recv);
	}
	return 0;
}
